<?php

namespace SGalinski\SgMail\Command;

/***************************************************************
 *  Copyright notice
 *
 *  (c) sgalinski Internet Services (https://www.sgalinski.de)
 *
 *  All rights reserved
 *
 *  This script is part of the TYPO3 project. The TYPO3 project is
 *  free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  The GNU General Public License can be found at
 *  http://www.gnu.org/copyleft/gpl.html.
 *
 *  This script is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  This copyright notice MUST APPEAR in all copies of the script!
 ***************************************************************/

use SGalinski\SgMail\Domain\Model\Mail;
use SGalinski\SgMail\Domain\Repository\MailRepository;
use SGalinski\SgMail\Service\PlaintextService;
use Symfony\Component\Console\Input\InputArgument;
use Symfony\Component\Console\Input\InputInterface;
use Symfony\Component\Console\Output\OutputInterface;
use TYPO3\CMS\Core\Mail\MailMessage;
use TYPO3\CMS\Core\Resource\FileInterface;
use TYPO3\CMS\Core\Utility\GeneralUtility;
use Symfony\Component\Console\Command\Command;
use TYPO3\CMS\Core\Utility\VersionNumberUtility;
use TYPO3\CMS\Extbase\Object\ObjectManager;

/**
 * Command controller for the mailing feature
 */
class SendMailCommandController extends Command {
	/**
	 * Configure command
	 */
	public function configure() {
		$this->setDescription('Executes the sending of mails with a maximum of the given sentCount.')
			->addArgument('sendCount', InputArgument::OPTIONAL, 'Send count', 50);
	}

	/**
	 * Execute the command
	 *
	 * @param InputInterface $input
	 * @param OutputInterface $output
	 * @return int
	 * @throws \TYPO3\CMS\Extbase\Object\Exception
	 * @throws \TYPO3\CMS\Extbase\Persistence\Exception\IllegalObjectTypeException
	 * @throws \TYPO3\CMS\Extbase\Persistence\Exception\UnknownObjectException
	 */
	public function execute(InputInterface $input, OutputInterface $output) {
		$sendCount = $input->getArgument('sendCount');
		if (version_compare(VersionNumberUtility::getCurrentTypo3Version(), '10.4.0', '<')) {
			$objectManager = GeneralUtility::makeInstance(ObjectManager::class);
			/** @var MailRepository $mailRepository */
			$mailRepository = $objectManager->get(MailRepository::class);
		} else {
			$mailRepository = GeneralUtility::makeInstance(MailRepository::class);
		}

		$mailsToSend = $mailRepository->findMailsToSend($sendCount);
		foreach ($mailsToSend as $mailToSend) {
			/** @var Mail $mailToSend */
			$fromAddress = $mailToSend->getFromAddress();
			$toAddresses = trim($mailToSend->getToAddress());
			$addressesArray = GeneralUtility::trimExplode(',', $toAddresses, TRUE);
			if (\count($addressesArray) > 1) {
				$toAddresses = $addressesArray;
			}
			$ccAddresses = GeneralUtility::trimExplode(',', $mailToSend->getCcAddresses(), TRUE);
			$bccAddresses = GeneralUtility::trimExplode(',', $mailToSend->getBccAddresses(), TRUE);
			$mailSubject = $mailToSend->getMailSubject();
			$mailBody = $mailToSend->getMailBody();

			$mailToSend->setSendingTime(time());
			$mailToSend->setLastSendingTime(time());

			if (empty($fromAddress) || empty($toAddresses)) {
				continue;
			}
			$mailRepository->update($mailToSend);

			$mailMessage = GeneralUtility::makeInstance(MailMessage::class);
			$mailMessage->setFrom($fromAddress, $mailToSend->getFromName());
			$mailMessage->setTo($toAddresses);
			$mailMessage->setSubject($mailSubject);
			if (\count($ccAddresses)) {
				$mailMessage->setCc($ccAddresses);
			}

			if (\count($bccAddresses)) {
				$mailMessage->setBcc($bccAddresses);
			}

			$replyTo = $mailToSend->getReplyTo();
			if ($replyTo) {
				$mailMessage->setReplyTo($replyTo);
			}

			$plaintextService = GeneralUtility::makeInstance(PlaintextService::class);
			$plaintextBody = $plaintextService->makePlain($mailBody);
			if (version_compare(VersionNumberUtility::getCurrentTypo3Version(), '10.4.0', '<')) {
				$mailMessage->setBody($mailBody, 'text/html');
				$mailMessage->addPart($plaintextBody, 'text/plain');
			} else {
				$mailMessage->html($mailBody);
				$mailMessage->text($plaintextBody);
			}

			$attachments = $mailToSend->getAttachments();
			if ($attachments->count() > 0) {
				foreach ($attachments as $attachment) {
					try {
						/** @var FileInterface $file */
						$file = $attachment->getOriginalResource();
						if (!$file) {
							continue;
						}

						$file = $file->getOriginalFile();
						if (!$file) {
							continue;
						}
						if (version_compare(VersionNumberUtility::getCurrentTypo3Version(), '10.4.0', '<')) {
							$mailMessage->attach(
								\Swift_Attachment::newInstance(
									$file->getContents(), $file->getName(), $file->getMimeType()
								)
							);
						} else {
							$mailMessage->attach($file->getContents(), $file->getName(), $file->getMimeType());
						}
					} catch (\Exception $exception) {
						continue;
					}
				}
			}

			try {
				$mailMessage->send();
			} catch (\Exception $exception) {
				// Ignore that. We can't do much here in a misconfiguration case.
			}
		}

		// Important for command controllers that change data
		$mailRepository->persist();
		return 0;
	}
}
