<?php

namespace SGalinski\SgMail\Domain\Repository;

/***************************************************************
 *  Copyright notice
 *
 *  (c) sgalinski Internet Services (https://www.sgalinski.de)
 *
 *  All rights reserved
 *
 *  This script is part of the TYPO3 project. The TYPO3 project is
 *  free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  The GNU General Public License can be found at
 *  http://www.gnu.org/copyleft/gpl.html.
 *
 *  This script is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  This copyright notice MUST APPEAR in all copies of the script!
 ***************************************************************/

use SGalinski\SgMail\Service\BackendService;
use TYPO3\CMS\Core\Database\ConnectionPool;
use TYPO3\CMS\Core\Utility\GeneralUtility;
use TYPO3\CMS\Extbase\Persistence\Generic\Query;
use TYPO3\CMS\Extbase\Persistence\QueryResultInterface;

/**
 * Repository for the Mail object
 */
class MailRepository extends AbstractRepository {
	const SENT = '1';
	const NOT_SENT = '2';

	/**
	 * Returns all mails that are still not sent ordered by priority.
	 *
	 * @param int $limit
	 * @return array|QueryResultInterface
	 */
	public function findMailsToSend(int $limit = NULL) {
		$query = $this->createQuery();
		if ($limit) {
			$query->setLimit($limit);
		}

		$query->setOrderings(['priority' => Query::ORDER_DESCENDING]);

		// ignore blacklisted templates
		$constraintsAnd[] = $query->equals('blacklisted', FALSE);
		$constraintsAnd[] = $query->equals('sending_time', 0);
		return $query->matching($query->logicalAnd($constraintsAnd))->execute();
	}

	/**
	 * Find all mails (sent & unsent) for extension key and template name
	 * We keep the ExtBase function only because of the pagination
	 *
	 * @param int $pid
	 * @param int $limit
	 * @param array $filters
	 * @return array|QueryResultInterface
	 * @throws \InvalidArgumentException
	 * @throws \TYPO3\CMS\Extbase\Persistence\Exception\InvalidQueryException
	 */
	public function findAllEntries($pid = 0, $limit = NULL, array $filters = []) {
		$query = $this->createQuery();
		if ($limit) {
			$query->setLimit($limit);
		}

		if ($filters['filterExtension'] && $filters['filterExtension'] !== 0
			&& $filters['filterExtension'] !== '0' && $filters['filterExtension'] !== ''
		) {
			$constraintsAnd[] = $query->equals('extension_key', $filters['filterExtension']);
		}

		if ($filters['filterTemplate'] && $filters['filterTemplate'] !== 0
			&& $filters['filterTemplate'] !== '0' && $filters['filterTemplate'] !== ''
		) {
			$constraintsAnd[] = $query->equals('template_name', $filters['filterTemplate']);
		}

		// if this language is no in the sys lang table, then include it with the default lang
		if ($filters['filterLanguage'] === 'default') {
			foreach (BackendService::getLanguages() as $lang) {
				// Exclude all other system languages than default
				if ($lang['isocode'] !== 'default') {
					$constraintsAnd[] = $query->logicalNot($query->equals('language', $lang['isocode']));
				}
			}
		} elseif ($filters['filterLanguage'] && $filters['filterLanguage'] !== 0
			&& $filters['filterLanguage'] !== '0' && $filters['filterLanguage'] !== ''
		) {
			$constraintsAnd[] = $query->equals('language', $filters['filterLanguage']);
		}

		$constraintsOr = [];
		if ($filters['filterSearch']) {
			$search = '%' . $filters['filterSearch'] . '%';
			if ($filters['filterFields'] && trim('' !== $filters['filterFields'])) {
				foreach ($filters['filterFields'] as $field) {
					switch ($field) {
						case BackendService::SENDER_FILTER_OPTION :
							$constraintsOr[] = $query->like('from_address', $search);
							break;
						case BackendService::RECIPIENT_FILTER_OPTION :
							$constraintsOr[] = $query->like('to_address', $search);
							break;
						case BackendService::SUBJECT_FILTER_OPTION :
							$constraintsOr[] = $query->like('mail_subject', $search);
							break;
						case BackendService::MAILTEXT_FILTER_OPTION :
							$constraintsOr[] = $query->like('mail_body', $search);
							break;
						case BackendService::CC_FILTER_OPTION :
							$constraintsOr[] = $query->like('cc_addresses', $search);
							break;
						case BackendService::BCC_FILTER_OPTION :
							$constraintsOr[] = $query->like('bcc_addresses', $search);
							break;
						case BackendService::FROM_NAME_FILTER_OPTION :
							$constraintsOr[] = $query->like('from_name', $search);
							break;
						case BackendService::REPLY_TO_NAME_FILTER_OPTION :
							$constraintsOr[] = $query->like('reply_to', $search);
							break;
					}
				}
			} else { // if nothing selected, search in all fields
				$constraintsOr[] = $query->like('from_address', $search);
				$constraintsOr[] = $query->like('to_address', $search);
				$constraintsOr[] = $query->like('mail_subject', $search);
				$constraintsOr[] = $query->like('mail_body', $search);
				$constraintsOr[] = $query->like('cc_addresses', $search);
				$constraintsOr[] = $query->like('bcc_addresses', $search);
				$constraintsOr[] = $query->like('from_name', $search);
				$constraintsOr[] = $query->like('reply_to', $search);
			}
		}
		if (count($constraintsOr)) {
			$constraintsAnd[] = $query->logicalOr($constraintsOr);
		}

		$fromTime = strtotime($filters['filterFromDate']);
		if ($fromTime) {
			$constraintsAnd[] = $query->greaterThanOrEqual('last_sending_time', $fromTime);
		}

		$toTime = strtotime($filters['filterToDate']);
		if ($toTime) {
			$constraintsAnd[] = $query->lessThanOrEqual('last_sending_time', $toTime);
		}

		if ((int) $filters['filterBlacklist'] !== 1) {
			$constraintsAnd[] = $query->equals('blacklisted', 0);
		}

		if ($filters['filterSent']) {
			switch ($filters['filterSent']) {
				case self::SENT:
					$constraintsAnd[] = $query->greaterThan('last_sending_time', 0);
					break;
				case self::NOT_SENT:
					$constraintsAnd[] = $query->equals('last_sending_time', 0);
					break;
			}
		}

		$constraintsAnd[] = $query->equals('pid', (int) $pid);

		$query->setOrderings(['tstamp' => Query::ORDER_DESCENDING]);
		$query->matching($query->logicalAnd($constraintsAnd));
		return $query->execute();
	}

	/**
	 * Find all mails (sent & unsent) for extension key and template name
	 * We use Doctrine for better performance
	 *
	 * @param int $pid
	 * @param int $limit
	 * @param array $filters
	 * @return \TYPO3\CMS\Core\Database\Query\QueryBuilder
	 * @throws \InvalidArgumentException
	 * @throws \TYPO3\CMS\Extbase\Persistence\Exception\InvalidQueryException
	 */
	public function getDoctrineQueryForFindAllEntries($pid = 0, $limit = NULL, array $filters = []) {
		$queryBuilder = GeneralUtility::makeInstance(ConnectionPool::class)->getQueryBuilderForTable(
			'tx_sgmail_domain_model_mail'
		)
			->from('tx_sgmail_domain_model_mail');
		$queryBuilder->where($queryBuilder->expr()->eq('deleted', 0));
		if ($limit) {
			$queryBuilder->setMaxResults($limit);
		}

		if ($filters['filterExtension'] && $filters['filterExtension'] !== 0
			&& $filters['filterExtension'] !== '0' && $filters['filterExtension'] !== ''
		) {
			$queryBuilder->andWhere($queryBuilder->expr()->eq('extension_key', $queryBuilder->createNamedParameter($filters['filterExtension'])));
		}

		if ($filters['filterTemplate'] && $filters['filterTemplate'] !== 0
			&& $filters['filterTemplate'] !== '0' && $filters['filterTemplate'] !== ''
		) {
			$queryBuilder->andWhere($queryBuilder->expr()->eq('template_name', $queryBuilder->createNamedParameter($filters['filterTemplate'])));
		}

		// if this language is no in the sys lang table, then include it with the default lang
		if ($filters['filterLanguage'] === 'default') {
			foreach (BackendService::getLanguages() as $lang) {
				// Exclude all other system languages than default
				if ($lang['isocode'] !== 'default') {
					$queryBuilder->andWhere($queryBuilder->expr()->neq('language', $queryBuilder->createNamedParameter($lang['isocode'])));
				}
			}
		} elseif ($filters['filterLanguage'] && $filters['filterLanguage'] !== 0
			&& $filters['filterLanguage'] !== '0' && $filters['filterLanguage'] !== ''
		) {
			$queryBuilder->andWhere($queryBuilder->expr()->eq('language', $queryBuilder->createNamedParameter($filters['filterLanguage'])));
		}

		$constraintsOr = [];
		if ($filters['filterSearch']) {
			$search = $queryBuilder->createNamedParameter('%' . $filters['filterSearch'] . '%');
			if ($filters['filterFields'] && trim('' !== $filters['filterFields'])) {
				foreach ($filters['filterFields'] as $field) {
					switch ($field) {
						case BackendService::SENDER_FILTER_OPTION :
							$constraintsOr[] = $queryBuilder->expr()->like('from_address', $search);
							break;
						case BackendService::RECIPIENT_FILTER_OPTION :
							$constraintsOr[] = $queryBuilder->expr()->like('to_address', $search);
							break;
						case BackendService::SUBJECT_FILTER_OPTION :
							$constraintsOr[] = $queryBuilder->expr()->like('mail_subject', $search);
							break;
						case BackendService::MAILTEXT_FILTER_OPTION :
							$constraintsOr[] = $queryBuilder->expr()->like('mail_body', $search);
							break;
						case BackendService::CC_FILTER_OPTION :
							$constraintsOr[] = $queryBuilder->expr()->like('cc_addresses', $search);
							break;
						case BackendService::BCC_FILTER_OPTION :
							$constraintsOr[] = $queryBuilder->expr()->like('bcc_addresses', $search);
							break;
						case BackendService::FROM_NAME_FILTER_OPTION :
							$constraintsOr[] = $queryBuilder->expr()->like('from_name', $search);
							break;
						case BackendService::REPLY_TO_NAME_FILTER_OPTION :
							$constraintsOr[] = $queryBuilder->expr()->like('reply_to', $search);
							break;
					}
				}
			} else { // if nothing selected, search in all fields
				$constraintsOr[] = $queryBuilder->expr()->like('from_address', $search);
				$constraintsOr[] = $queryBuilder->expr()->like('to_address', $search);
				$constraintsOr[] = $queryBuilder->expr()->like('mail_subject', $search);
				$constraintsOr[] = $queryBuilder->expr()->like('mail_body', $search);
				$constraintsOr[] = $queryBuilder->expr()->like('cc_addresses', $search);
				$constraintsOr[] = $queryBuilder->expr()->like('bcc_addresses', $search);
				$constraintsOr[] = $queryBuilder->expr()->like('from_name', $search);
				$constraintsOr[] = $queryBuilder->expr()->like('reply_to', $search);
			}
		}
		if (count($constraintsOr)) {
			$queryBuilder->andWhere($queryBuilder->expr()->orX(... $constraintsOr));
		}

		$fromTime = strtotime($filters['filterFromDate']);
		if ($fromTime) {
			$queryBuilder->andWhere($queryBuilder->expr()->gte('last_sending_time', $queryBuilder->createNamedParameter($fromTime)));
		}

		$toTime = strtotime($filters['filterToDate']);
		if ($toTime) {
			$queryBuilder->andWhere($queryBuilder->expr()->lte('last_sending_time', $queryBuilder->createNamedParameter($toTime)));
		}

		if ((int) $filters['filterBlacklist'] !== 1) {
			$queryBuilder->andWhere($queryBuilder->expr()->eq('blacklisted', 0));
		}

		if ($filters['filterSent']) {
			switch ($filters['filterSent']) {
				case self::SENT:
					$queryBuilder->andWhere($queryBuilder->expr()->gt('last_sending_time', 0));
					break;
				case self::NOT_SENT:
					$queryBuilder->andWhere($queryBuilder->expr()->eq('last_sending_time', 0));
					break;
			}
		}

		$queryBuilder->andWhere($queryBuilder->expr()->eq('pid', (int) $pid));

		return $queryBuilder->orderBy('tstamp', 'desc');
	}
}
